@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2022 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -23,6 +23,8 @@
 #include "gxcomp.h"
 #include "gsmatrix.h"
 #include "gxblend.h"
+#include "gdevp14.h"
+#include "gsfunc.h"
 
 /*
  * Define the operations for the PDF 1.4 transparency compositor.
@@ -33,6 +35,7 @@ typedef enum {
     PDF14_ABORT_DEVICE,
     PDF14_BEGIN_TRANS_GROUP,
     PDF14_END_TRANS_GROUP,
+    PDF14_BEGIN_TRANS_PAGE_GROUP,
     PDF14_BEGIN_TRANS_TEXT_GROUP,
     PDF14_END_TRANS_TEXT_GROUP,
     PDF14_BEGIN_TRANS_MASK,
@@ -51,6 +54,7 @@ typedef enum {
     "PDF14_ABORT_DEVICE     ",\
     "PDF14_BEGIN_TRANS_GROUP",\
     "PDF14_END_TRANS_GROUP  ",\
+    "PDF14_BEGIN_TRANS_PAGE_GROUP",\
     "PDF14_BEGIN_TRANS_TEXT_GROUP  ",\
     "PDF14_END_TRANS_TEXT_GROUP  ",\
     "PDF14_BEGIN_TRANS_MASK ",\
@@ -65,10 +69,12 @@ typedef enum {
 /* Bit definitions for serializing PDF 1.4 parameters */
 #define PDF14_SET_BLEND_MODE    (1 << 0)
 #define PDF14_SET_TEXT_KNOCKOUT (1 << 1)
-#define PDF14_SET_SHAPE_ALPHA   (1 << 2)
-#define PDF14_SET_OPACITY_ALPHA (1 << 3)
+#define PDF14_SET_AIS   (1 << 2)
 #define PDF14_SET_OVERPRINT		(1 << 4)
-#define PDF14_SET_OVERPRINT_MODE (1 << 5)
+#define PDF14_SET_FILLCONSTANTALPHA (1 << 6)
+#define PDF14_SET_STROKECONSTANTALPHA (1 << 7)
+#define PDF14_SET_STROKEOVERPRINT (1 << 8)
+#define PDF14_SET_FILLSTROKE_STATE (1 << 9)
 
 /* Used for keeping track of the text group madness, since we have the pdf14
    device needs to know if we are int an BT/ET group vs. a FreeText Annotation
@@ -77,22 +83,16 @@ typedef enum {
 typedef enum {
     PDF14_TEXTGROUP_NO_BT,  /* We are not in a BT/ET.  Avoids Annotation Texts */
     PDF14_TEXTGROUP_BT_NOT_PUSHED, /* We are in a BT/ET but no group pushed */
-    PDF14_TEXTGROUP_BT_PUSHED   /* We are in a BT/ET section and group was pushed */
+    PDF14_TEXTGROUP_BT_PUSHED,   /* We are in a BT/ET section and group was pushed */
+    PDF14_TEXTGROUP_MISSING_ET   /* We pushed a group already and then had another BT occur */
 } pdf14_text_group_state;
 
-#ifndef gs_function_DEFINED
-typedef struct gs_function_s gs_function_t;
-#  define gs_function_DEFINED
-#endif
-
-typedef struct gs_transparency_source_s {
-    float alpha;		/* constant alpha */
-} gs_transparency_source_t;
-
 struct gs_pdf14trans_params_s {
     /* The type of trasnparency operation */
     pdf14_compositor_operations pdf14_op;
     int num_spot_colors;    /* Only for devices which support spot colors. */
+    int num_spot_colors_int; /* Number of spot colors reported by the interpreter */
+    bool overprint_sim_push; /* If true, put_image will need special conversion */
     /* Changed parameters flag */
     int changed;
     /* Parameters from the gs_transparency_group_params_t structure */
@@ -107,6 +107,7 @@ struct gs_pdf14trans_params_s {
     bool function_is_identity;
     int Background_components;
     float Background[GS_CLIENT_COLOR_MAX_COMPONENTS];
+    const gs_color_space *ColorSpace;
     int Matte_components;
     float Matte[GS_CLIENT_COLOR_MAX_COMPONENTS];
     float GrayBackground;  /* This is used to determine if the
@@ -116,34 +117,37 @@ struct gs_pdf14trans_params_s {
                               outside its own groups bounding
                               box in such a case */
     gs_function_t *transfer_function;
-    byte transfer_fn[MASK_TRANSFER_FUNCTION_SIZE];
+    byte transfer_fn[MASK_TRANSFER_FUNCTION_SIZE*2+2];
     /* Individual transparency parameters */
     gs_blend_mode_t blend_mode;
     bool text_knockout;
     int text_group;
-    gs_transparency_source_t opacity;
-    gs_transparency_source_t shape;
+    bool page_group;
+    bool ais;
+    bool shade_group;
+    float opacity;
+    float shape;
+    float fillconstantalpha;
+    float strokeconstantalpha;
     bool mask_is_image;
     gs_matrix ctm;
     bool replacing;
     bool overprint;
-    bool overprint_mode;
+    bool stroke_overprint;
+    bool effective_overprint_mode;
+    bool stroke_effective_op_mode;
     bool idle; /* For clist reader.*/
     uint mask_id; /* For clist reader.*/
     int group_color_numcomps;
-    gs_transparency_color_t group_color;
+    gs_transparency_color_t group_color_type;
     int64_t icc_hash;
     cmm_profile_t *iccprofile;               /* The profile  */
     bool crop_blend_params;  /* This is used when the blend params are updated
                                 during a transparency group push */
     bool is_pattern;      /* Needed to detect device push and pop for clist pattern */
+    PDF14_OP_FS_STATE op_fs_state;
 };
 
-#ifndef gs_pdf14trans_params_DEFINED
-#define gs_pdf14trans_params_DEFINED
-typedef struct gs_pdf14trans_params_s gs_pdf14trans_params_t;
-#endif
-
 /*
  * The PDF 1.4 transparency compositor structure. This is exactly analogous to
  * other compositor structures, consisting of a the compositor common elements
@@ -157,10 +161,6 @@ typedef struct gs_pdf14trans_s {
 /* Access transparency-related graphics state elements. */
 int gs_setblendmode(gs_gstate *, gs_blend_mode_t);
 gs_blend_mode_t gs_currentblendmode(const gs_gstate *);
-int gs_setopacityalpha(gs_gstate *, double);
-float gs_currentopacityalpha(const gs_gstate *);
-int gs_setshapealpha(gs_gstate *, double);
-float gs_currentshapealpha(const gs_gstate *);
 int gs_settextknockout(gs_gstate *, bool);
 bool gs_currenttextknockout(const gs_gstate *);
 
@@ -168,19 +168,29 @@ bool gs_currenttextknockout(const gs_gstate *);
  * We have to abbreviate the procedure name because procedure names are
  * only unique to 23 characters on VMS.
  */
-int gs_push_pdf14trans_device(gs_gstate * pgs, bool is_pattern);
+/* For the push_pdf14trans_device, depth is a estimate (not authoritative) of
+ * the transparency stack depth. Value 0 means "unknown". If the depth value
+ * is < 0, that means that this pdf14 device is being used for the overprint
+ * simulation, and ordiniarly the value will be -1, since overprint
+ * simulation only needs the page buffer.
+ * The spot_color_count is provided for PDF input files to define the number
+ * of Spot colorants beyond CMYK. Thus 0 means CMYK only. If unknown, such
+ * as with PostScript input, the value will be -1.
+ */
+int gs_push_pdf14trans_device(gs_gstate * pgs, bool is_pattern, bool retain,
+                              int depth, int spot_color_count);
 
 int gs_pop_pdf14trans_device(gs_gstate * pgs, bool is_pattern);
 
 int gs_abort_pdf14trans_device(gs_gstate * pgs);
 
-void gs_trans_group_params_init(gs_transparency_group_params_t *ptgp);
+void gs_trans_group_params_init(gs_transparency_group_params_t *ptgp, float opacity);
 
 int gs_update_trans_marking_params(gs_gstate * pgs);
 
 int gs_begin_transparency_group(gs_gstate * pgs,
                                 const gs_transparency_group_params_t *ptgp,
-                                const gs_rect *pbbox);
+                                const gs_rect *pbbox, pdf14_compositor_operations group_type);
 
 int gs_end_transparency_group(gs_gstate *pgs);
 
@@ -257,12 +267,11 @@ bool gs_getalphaisshape(gs_gstate *pgs);
 #define NUM_PDF14_BUFFERS 4     /* totally a random guess */
 #define NUM_ALPHA_CHANNELS 1    /* common, but doesn't include possible tag, shape or group alpha */
 #define NUM_COLOR_CHANNELS 4    /* CMYK is most common 'worst case' */
-#define BITS_PER_CHANNEL 8      /* currently pdf14 device is always 8-bit */
 /* The estimated size of an individual PDF 1.4 buffer row (in bits) */
-#define ESTIMATED_PDF14_ROW_SIZE(width, target_num_components) ((width) * BITS_PER_CHANNEL\
+#define ESTIMATED_PDF14_ROW_SIZE(width, target_num_components, bits) ((width) * (bits)\
         * (NUM_ALPHA_CHANNELS + max(target_num_components,NUM_COLOR_CHANNELS)))
 /* The estimated size of one row in all PDF 1.4 buffers (in bits) */
-#define ESTIMATED_PDF14_ROW_SPACE(width, target_num_components) \
-        (NUM_PDF14_BUFFERS * ESTIMATED_PDF14_ROW_SIZE(width, target_num_components))
+#define ESTIMATED_PDF14_ROW_SPACE(width, target_num_components, bits) \
+        (NUM_PDF14_BUFFERS * ESTIMATED_PDF14_ROW_SIZE(width, target_num_components, bits))
 
 #endif /* gstrans_INCLUDED */
