@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -71,33 +71,71 @@ clist_bitmap_bytes(uint width_bits, uint height, int compression_mask,
  * necessary.  We require height >= 1, raster >= bitmap_raster(width_bits).
  */
 static int
+go_process(stream_state * st, stream_cursor_read *pr, stream_cursor_write *pw, bool end)
+{
+    int status = (*st->templat->process) (st, pr, pw, end);
+    if (status)
+        return status;
+    /* We don't attempt to handle compressors that */
+    /* require >1 input byte to make progress. */
+    if (pr->ptr != pr->limit)
+        return -1;
+    return 0;
+}
+static byte zeros[1<<align_bitmap_mod] = { 0, };
+static int
 cmd_compress_bitmap(stream_state * st, const byte * data, uint width_bits,
                     uint raster, uint height, stream_cursor_write * pw)
 {
     uint width_bytes = bitmap_raster(width_bits);
     int status = 0;
     stream_cursor_read r;
+    stream_cursor_read r2;
+    uint whole_bytes = width_bits>>3;
+    uint mask = (0xff00>>(width_bits & 7)) & 0xff;
+    uint padding = width_bytes - ((width_bits+7)>>3);
 
-    r.ptr = data - 1;
-    if (raster == width_bytes) {
-        r.limit = r.ptr + raster * height;
+    if (raster == whole_bytes) {
+        stream_cursor_read_init(&r, data, raster * (size_t)height);
         status = (*st->templat->process) (st, &r, pw, true);
     } else {			/* Compress row-by-row. */
         uint y;
 
-        for (y = 1; (r.limit = r.ptr + width_bytes), y < height; ++y) {
-            status = (*st->templat->process) (st, &r, pw, false);
+        stream_cursor_read_init(&r, data, whole_bytes);
+
+        for (y = height-1; (r.limit = r.ptr + whole_bytes), y > 0; y--) {
+            status = go_process(st, &r, pw, false);
             if (status)
                 break;
-            if (r.ptr != r.limit) {	/* We don't attempt to handle compressors that */
-                /* require >1 input byte to make progress. */
-                status = -1;
-                break;
+            if (mask) {
+                byte b = r.ptr[1] & mask;
+
+                stream_cursor_read_init(&r2, &b, 1);
+                status = go_process(st, &r2, pw, false);
+                if (status)
+                    break;
+            }
+            if (padding) {
+                stream_cursor_read_init(&r2, zeros, padding);
+                status = go_process(st, &r2, pw, false);
+                if (status)
+                    break;
+            }
+            r.ptr += (int)(raster - whole_bytes);
+        }
+        if (status == 0) {
+            status = go_process(st, &r, pw, padding == 0 && mask == 0);
+            if (status == 0 && mask) {
+                byte b = r.ptr[1] & mask;
+
+                stream_cursor_read_init(&r2, &b, 1);
+                status = go_process(st, &r2, pw, padding == 0);
+            }
+            if (status == 0 && padding) {
+                stream_cursor_read_init(&r2, zeros, padding);
+                status = go_process(st, &r2, pw, true);
             }
-            r.ptr += raster - width_bytes;
         }
-        if (status == 0)
-            status = (*st->templat->process) (st, &r, pw, true);
     }
     if (st->templat->release)
         (*st->templat->release) (st);
@@ -110,6 +148,7 @@ cmd_compress_bitmap(stream_state * st, const byte * data, uint width_bits,
  * Return <0 if error, otherwise the compression method.
  * A return value of gs_error_limitcheck means that the bitmap was too big
  * to fit in the command reading buffer.
+ * This won't happen if the compression_mask has allow_large_bitmap set.
  * Note that this leaves room for the command and initial arguments,
  * but doesn't fill them in.
  */
@@ -119,15 +158,14 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
              int compression_mask, byte ** pdp, uint * psize)
 {
     uint short_raster, full_raster;
-    uint short_size =
-    clist_bitmap_bytes(width_bits, height,
-                       compression_mask & ~cmd_mask_compress_any,
-                       &short_raster, &full_raster);
+    uint short_size = clist_bitmap_bytes(width_bits, height,
+                          compression_mask & ~cmd_mask_compress_any,
+                          &short_raster, &full_raster);
     uint uncompressed_raster;
-    uint uncompressed_size =
-    clist_bitmap_bytes(width_bits, height, compression_mask,
+    uint uncompressed_size = clist_bitmap_bytes(width_bits, height, compression_mask,
                        &uncompressed_raster, &full_raster);
-    uint max_size = data_bits_size - op_size;
+    uint max_size = (compression_mask & allow_large_bitmap) ? 0x7fffffff :
+                        data_bits_size - op_size;
     gs_memory_t *mem = cldev->memory;
     byte *dp;
     int compress = 0;
@@ -153,8 +191,8 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
 
         *psize = try_size;
         code = (pcls != 0 ?
-                set_cmd_put_op(dp, cldev, pcls, 0, try_size) :
-                set_cmd_put_all_op(dp, cldev, 0, try_size));
+                set_cmd_put_op(&dp, cldev, pcls, 0, try_size) :
+                set_cmd_put_all_op(&dp, cldev, 0, try_size));
         if (code < 0)
             return code;
         cmd_uncount_op(0, try_size);
@@ -189,13 +227,13 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
             w.ptr = wbase;
             w.limit = w.ptr + min(wmax, short_size >> 1);
             status = cmd_compress_bitmap((stream_state *) & sstate, data,
-                                  uncompressed_raster << 3 /*width_bits */ ,
+                                  width_bits, /* was uncompressed_raster << 3, but this overruns. */
                                          raster, height, &w);
             if (status == 0) {	/* Use compressed representation. */
                 uint wcount = w.ptr - wbase;
 
                 cmd_shorten_list_op(cldev,
-                             (pcls ? &pcls->list : &cldev->band_range_list),
+                             (pcls ? &pcls->list : cldev->band_range_list),
                                     try_size - (op_size + wcount));
                 *psize = op_size + wcount;
                 goto out;
@@ -207,7 +245,7 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
                        "[L]Uncompressed bits %u too large for buffer\n",
                        uncompressed_size);
             cmd_shorten_list_op(cldev,
-                             (pcls ? &pcls->list : &cldev->band_range_list),
+                             (pcls ? &pcls->list : cldev->band_range_list),
                                 try_size);
             return_error(gs_error_limitcheck);
         }
@@ -215,7 +253,7 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
             if_debug2m('L',cldev->memory,"[L]Shortening bits from %u to %u\n",
                        try_size, op_size + short_size);
             cmd_shorten_list_op(cldev,
-                             (pcls ? &pcls->list : &cldev->band_range_list),
+                             (pcls ? &pcls->list : cldev->band_range_list),
                                 try_size - (op_size + short_size));
             *psize = op_size + short_size;
         }
@@ -225,8 +263,8 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
     else {
         *psize = op_size + short_size;
         code = (pcls != 0 ?
-                set_cmd_put_op(dp, cldev, pcls, 0, *psize) :
-                set_cmd_put_all_op(dp, cldev, 0, *psize));
+                set_cmd_put_op(&dp, cldev, pcls, 0, *psize) :
+                set_cmd_put_all_op(&dp, cldev, 0, *psize));
         if (code < 0)
             return code;
         cmd_uncount_op(0, *psize);
@@ -234,14 +272,15 @@ cmd_put_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
     if ((compression_mask & (1 << cmd_compress_const)) &&
         (code = bytes_rectangle_is_const(data, raster, uncompressed_raster << 3, height)) >= 0) {
         cmd_shorten_list_op(cldev,
-                            (pcls ? &pcls->list : &cldev->band_range_list),
+                            (pcls ? &pcls->list : cldev->band_range_list),
                             *psize - (op_size + 1));
         *psize = op_size + 1;
         dp[op_size] = code;
         compress = cmd_compress_const;
     } else {
         uint copy_bytes = (width_bits + 7) >> 3;
-        bytes_copy_rectangle_zero_padding(dp + op_size, short_raster, data, raster,
+        bytes_copy_rectangle_zero_padding_last_short(
+                             dp + op_size, short_raster, data, raster,
                              copy_bytes, height);
     }
 out:
@@ -304,12 +343,12 @@ cmd_put_tile_index(gx_device_clist_writer *cldev, gx_clist_state *pcls,
     int code;
 
     if (!(idelta & ~15)) {
-        code = set_cmd_put_op(dp, cldev, pcls,
+        code = set_cmd_put_op(&dp, cldev, pcls,
                               cmd_op_delta_tile_index + idelta, 1);
         if (code < 0)
             return code;
     } else {
-        code = set_cmd_put_op(dp, cldev, pcls,
+        code = set_cmd_put_op(&dp, cldev, pcls,
                               cmd_op_set_tile_index + (indx >> 8), 2);
         if (code < 0)
             return code;
@@ -331,7 +370,7 @@ cmd_put_color_map(gx_device_clist_writer * cldev, cmd_map_index map_index,
     if (map == 0) {
         if (pid && *pid == gs_no_id)
             return 0;	/* no need to write */
-        code = set_cmd_put_all_op(dp, cldev, cmd_opv_set_misc, 3);
+        code = set_cmd_put_all_op(&dp, cldev, cmd_opv_set_misc, 3);
         if (code < 0)
             return code;
         dp[1] = cmd_set_misc_map + (cmd_map_none << 4) + map_index;
@@ -342,13 +381,13 @@ cmd_put_color_map(gx_device_clist_writer * cldev, cmd_map_index map_index,
         if (pid && map->id == *pid)
             return 0;	/* no need to write */
         if (map->proc == gs_identity_transfer) {
-            code = set_cmd_put_all_op(dp, cldev, cmd_opv_set_misc, 3);
+            code = set_cmd_put_all_op(&dp, cldev, cmd_opv_set_misc, 3);
             if (code < 0)
                 return code;
             dp[1] = cmd_set_misc_map + (cmd_map_identity << 4) + map_index;
             dp[2] = comp_num;
         } else {
-            code = set_cmd_put_all_op(dp, cldev, cmd_opv_set_misc,
+            code = set_cmd_put_all_op(&dp, cldev, cmd_opv_set_misc,
                                       3 + sizeof(map->values));
             if (code < 0)
                 return code;
@@ -448,21 +487,19 @@ clist_add_tile(gx_device_clist_writer * cldev, const gx_strip_bitmap * tiles,
     uint size_bytes = raster * tiles->size.y * tiles->num_planes;
     uint tsize =
     sizeof(tile_slot) + cldev->tile_band_mask_size + size_bytes;
-    gx_cached_bits_head *slot_head;
-
-#define slot ((tile_slot *)slot_head)
+    tile_slot *slot;
 
     if (cldev->bits.csize == cldev->tile_max_count) {	/* Don't let the hash table get too full: delete an entry. */
         /* Since gx_bits_cache_alloc returns an entry to delete when */
         /* it fails, just force it to fail. */
         gx_bits_cache_alloc(&cldev->bits, (ulong) cldev->cache_chunk->size,
-                            &slot_head);
-        if (slot_head == 0) {	/* Wrap around and retry. */
+                            (gx_cached_bits_head **)&slot);
+        if (slot == NULL) {	/* Wrap around and retry. */
             cldev->bits.cnext = 0;
             gx_bits_cache_alloc(&cldev->bits, (ulong) cldev->cache_chunk->size,
-                                &slot_head);
+                                (gx_cached_bits_head **)&slot);
 #ifdef DEBUG
-            if (slot_head == 0) {
+            if (slot == NULL) {
                 lprintf("No entry to delete!\n");
                 return_error(gs_error_Fatal);
             }
@@ -471,8 +508,8 @@ clist_add_tile(gx_device_clist_writer * cldev, const gx_strip_bitmap * tiles,
         clist_delete_tile(cldev, slot);
     }
     /* Allocate the space for the new entry, deleting entries as needed. */
-    while (gx_bits_cache_alloc(&cldev->bits, (ulong) tsize, &slot_head) < 0) {
-        if (slot_head == 0) {	/* Wrap around. */
+    while (gx_bits_cache_alloc(&cldev->bits, (ulong) tsize, (gx_cached_bits_head **)&slot) < 0) {
+        if (slot == NULL) {	/* Wrap around. */
             if (cldev->bits.cnext == 0) {	/* Too big to fit.  We should probably detect this */
                 /* sooner, since if we get here, we've cleared the */
                 /* cache. */
@@ -483,8 +520,8 @@ clist_add_tile(gx_device_clist_writer * cldev, const gx_strip_bitmap * tiles,
             clist_delete_tile(cldev, slot);
     }
     /* Fill in the entry. */
-    slot->cb_depth = depth;
-    slot->cb_raster = raster;
+    slot->head.depth = depth;
+    slot->raster = raster;
     slot->width = tiles->rep_width;
     slot->height = tiles->rep_height;
     slot->shift = slot->rep_shift = tiles->rep_shift;
@@ -511,7 +548,7 @@ clist_add_tile(gx_device_clist_writer * cldev, const gx_strip_bitmap * tiles,
 #endif
         slot->index = loc.index;
         cldev->tile_table[loc.index].offset =
-            (byte *) slot_head - cldev->data;
+            (byte *) slot - cldev->data;
         if_debug2m('L', cldev->memory, "[L]adding index=%u, offset=%lu\n",
                    loc.index, cldev->tile_table[loc.index].offset);
     }
@@ -659,18 +696,22 @@ clist_change_tile(gx_device_clist_writer * cldev, gx_clist_state * pcls,
                 if (tiles->num_planes != 1)
                     pdepth /= tiles->num_planes;
 
+                /* put the bits, but don't restrict to a single buffer */
                 code = cmd_put_bits(cldev, pcls, ts_bits(cldev, loc.tile),
                                     tiles->rep_width * pdepth,
                                     tiles->rep_height * tiles->num_planes,
-                                    loc.tile->cb_raster, rsize,
-                                    (cldev->tile_params.size.x > tiles->rep_width ?
-                                     decompress_elsewhere | decompress_spread :
-                                     decompress_elsewhere),
+                                    loc.tile->raster, rsize,
+                                    allow_large_bitmap |
+                                        (cldev->tile_params.size.x > tiles->rep_width ?
+                                             decompress_elsewhere | decompress_spread :
+                                             decompress_elsewhere),
                                     &dp, &csize);
 
                 if (code < 0)
                     return code;
                 if (extra) {	/* Write the tile parameters before writing the bits. */
+                    if_debug1m('L', cldev->memory,
+                               "[L] fake end_run: really set_tile_size[%d]\n", extra);
                     cmd_store_tile_params(dp, &cldev->tile_params, depth,
                                           extra, for_pattern, cldev->memory);
                     dp += extra;
@@ -681,6 +722,8 @@ clist_change_tile(gx_device_clist_writer * cldev, gx_clist_state * pcls,
                     if (band_index > cldev->tile_known_max)
                         cldev->tile_known_max = band_index;
                 }
+                if_debug1m('L', cldev->memory,
+                           "[L] fake end_run: really set_tile_bits[%d]\n", csize-extra);
                 *dp = cmd_count_op(cmd_opv_set_tile_bits, csize - extra, cldev->memory);
                 dp++;
                 dp = cmd_put_w(loc.index, dp);
@@ -722,66 +765,78 @@ clist_change_bits(gx_device_clist_writer * cldev, gx_clist_state * pcls,
 {
     tile_loc loc;
     int code;
+    uint band_index = pcls - cldev->states;
+    byte bmask = 1 << (band_index & 7);
+    byte *bptr;
 
-   top:if (clist_find_bits(cldev, tiles->id, &loc)) {	/* The bitmap is in the cache.  Check whether this band */
-        /* knows about it. */
-        uint band_index = pcls - cldev->states;
-        byte *bptr = ts_mask(loc.tile) + (band_index >> 3);
-        byte bmask = 1 << (band_index & 7);
-
-        if (*bptr & bmask) {	/* Already known.  Just set the index. */
-            if (pcls->tile_index == loc.index)
-                return 0;
-            cmd_put_tile_index(cldev, pcls, loc.index);
-        } else {		/* Not known yet.  Output the bits. */
-            /* Note that the offset we write is the one used by */
-            /* the reading phase, not the writing phase. */
-            ulong offset = (byte *) loc.tile - cldev->cache_chunk->data;
-            uint rsize = 2 + cmd_size_w(loc.tile->width) +
-            cmd_size_w(loc.tile->height) + cmd_size_w(loc.index) +
-            cmd_size_w(offset);
-            byte *dp;
-            uint csize;
-            gx_clist_state *bit_pcls = pcls;
-            int code;
-            int pdepth = depth;
-
-            if (tiles->num_planes != 1)
-                    pdepth /= loc.tile->num_planes;
-            if (loc.tile->num_bands == CHAR_ALL_BANDS_COUNT)
-                bit_pcls = NULL;
-            code = cmd_put_bits(cldev, bit_pcls, ts_bits(cldev, loc.tile),
-                                loc.tile->width * pdepth,
-                                loc.tile->height * loc.tile->num_planes, loc.tile->cb_raster,
-                                rsize,
-                              decompress_elsewhere | (cldev->target->BLS_force_memory ? (1 << cmd_compress_cfe) : 0),
-                                &dp, &csize);
-
-            if (code < 0)
-                return code;
-            *dp = cmd_count_op(cmd_opv_set_bits, csize, cldev->memory);
-            dp[1] = (depth << 2) + code;
-            dp += 2;
-            dp = cmd_put_w(loc.tile->width, dp);
-            dp = cmd_put_w(loc.tile->height, dp);
-            dp = cmd_put_w(loc.index, dp);
-            cmd_put_w(offset, dp);
-            if (bit_pcls == NULL) {
-                memset(ts_mask(loc.tile), 0xff,
-                       cldev->tile_band_mask_size);
-                loc.tile->num_bands = cldev->nbands;
-            } else {
-                *bptr |= bmask;
-                loc.tile->num_bands++;
-            }
-        }
-        pcls->tile_index = loc.index;
-        pcls->tile_id = loc.tile->id;
-        return 0;
+    while (!clist_find_bits(cldev, tiles->id, &loc)) {
+        /* The tile is not in the cache. */
+        code = clist_add_tile(cldev, tiles, tiles->raster, depth);
+        if (code < 0)
+            return code;
     }
-    /* The tile is not in the cache. */
-    code = clist_add_tile(cldev, tiles, tiles->raster, depth);
-    if (code < 0)
-        return code;
-    goto top;
+
+    /* The bitmap is in the cache.  Check whether this band */
+    /* knows about it. */
+    bptr = ts_mask(loc.tile) + (band_index >> 3);
+
+    if (*bptr & bmask) {	/* Already known.  Just set the index. */
+        if (pcls->tile_index == loc.index)
+            return 0;
+        cmd_put_tile_index(cldev, pcls, loc.index);
+    } else {		/* Not known yet.  Output the bits. */
+        /* Note that the offset we write is the one used by */
+        /* the reading phase, not the writing phase. */
+        ulong offset = (byte *) loc.tile - cldev->cache_chunk->data;
+        uint rsize = 2 + cmd_size_w(loc.tile->width) +
+                     cmd_size_w(loc.tile->height) +
+                     (loc.tile->num_planes > 1 ? 1 : 0) +
+                     cmd_size_w(loc.index) +
+                     cmd_size_w(offset);
+        byte *dp;
+        uint csize;
+        gx_clist_state *bit_pcls = pcls;
+        int pdepth = depth;
+
+        if (tiles->num_planes != 1)
+            pdepth /= loc.tile->num_planes;
+        if (loc.tile->num_bands == CHAR_ALL_BANDS_COUNT)
+            bit_pcls = NULL;
+        /* put the bits, but don't restrict to a single buffer */
+        code = cmd_put_bits(cldev, bit_pcls, ts_bits(cldev, loc.tile),
+                            loc.tile->width * pdepth,
+                            loc.tile->height * loc.tile->num_planes, loc.tile->raster,
+                            rsize,
+                            decompress_elsewhere |
+                                (cldev->target->BLS_force_memory ? (1 << cmd_compress_cfe) : 0),
+                            &dp, &csize);
+
+        if (code < 0)
+            return code;
+        if_debug1m('L', cldev->memory,
+                   "[L] fake end_run: really set_bits[%d]\n", csize);
+        *dp = cmd_count_op(loc.tile->num_planes > 1 ? cmd_opv_set_bits_planar : cmd_opv_set_bits,
+                           csize, cldev->memory);
+        dp[1] = (depth << 2) + code;
+        dp += 2;
+        dp = cmd_put_w(loc.tile->width, dp);
+        dp = cmd_put_w(loc.tile->height, dp);
+        if (loc.tile->num_planes > 1)
+            *dp++ = loc.tile->num_planes;
+        dp = cmd_put_w(loc.index, dp);
+        cmd_put_w(offset, dp);
+        if_debug7m('L', cldev->memory, " compress=%d depth=%d size=(%d,%d) planes=%d index=%d offset=%ld\n",
+                   code, depth, loc.tile->width, loc.tile->height, loc.tile->num_planes, loc.index, offset);
+        if (bit_pcls == NULL) {
+            memset(ts_mask(loc.tile), 0xff,
+                   cldev->tile_band_mask_size);
+            loc.tile->num_bands = cldev->nbands;
+        } else {
+            *bptr |= bmask;
+            loc.tile->num_bands++;
+        }
+    }
+    pcls->tile_index = loc.index;
+    pcls->tile_id = loc.tile->id;
+    return 0;
 }
